package jalview.rest;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.CompletableFuture;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import jalview.bin.Cache;
import jalview.rest.RestHandler.Status;

public abstract class EndpointAsync extends Endpoint
{
  public EndpointAsync(API api, final String name)
  {
    super(api, name);
  }

  protected String idExtension = null;

  protected String id = null;

  protected CompletableFuture<Void> cf = null;

  protected Map<String, CompletableFuture<Void>> cfMap = new HashMap<>();

  protected Map<String, String> stringMap = new HashMap<>();

  protected void setCompletableFuture(CompletableFuture<Void> cf)
  {
    this.cf = cf;
    if (getId() != null)
      cfMap.put(getId(), cf);
  }

  protected CompletableFuture<Void> getCompletableFuture()
  {
    if (cf == null && getId() != null && cfMap.get(getId()) != null)
      cf = cfMap.get(getId());
    return this.cf;
  }

  protected void setId(String id)
  {
    this.id = id;
  }

  protected void setIdExtension(String idExtension)
  {
    setId(getName() + "::" + idExtension);
  }

  protected String getId()
  {
    return this.id;
  }

  /*
   * Override the three methods
   * initialise (get parameters, set id (extension), set cf if using an existing one)
   * process (what to do in the cf if not using an existing one)
   * finalise (extra stuff to do at the end of the first call to this)
   */
  protected void initialise(HttpServletRequest request,
          HttpServletResponse response)
  {
    // should be overridden
    // must setId(request, extension)
    setId(request, null);
  }

  protected abstract void process(HttpServletRequest request,
          HttpServletResponse response, final Map<String, String> finalMap);

  protected void finalise(HttpServletRequest request,
          HttpServletResponse response)
  {
    // can be Overridden
  }

  @Override
  public void processEndpoint(HttpServletRequest request,
          HttpServletResponse response)
  {
    // subclass method
    Cache.info("**************REMOVEME about to run initialise()");
    initialise(request, response);

    Cache.info("**************REMOVEME about to run checkStatus()");
    if (checkStatus(request, response, Status.STARTED))
    {
      returnStatus(response);
      return;
    }

    Cache.info(
            "**************REMOVEME about to run getCompletableFuture()");
    if (getCompletableFuture() == null)
    {
      final Map<String, String> finalMap = stringMap;
      setCompletableFuture(CompletableFuture.runAsync(() -> {
        // subclass method
        this.process(request, response, finalMap);
      }));
    }
    else
    {
      Cache.info(
              "####################REMOVEME CompletableFuture cf is already set cf="
                      + cf);
    }
    Cache.info(
            "####################REMOVEME about to run finaliseCompletableFuture()");
    finaliseCompletableFuture();

    // subclass method
    finalise(request, response);

    Cache.info("REMOVEME about to run returnStatus");
    returnStatus(response);
    Cache.info("REMOVEME about to change status for id=" + getId());
    changeStatus(Status.IN_PROGRESS);
    Cache.info("REMOVEME now changed status for id=" + getId()
            + " to status=" + getStatus());
  }

  protected void atEnd()
  {
  }

  /*
   * Shared methods below here
   */

  protected String setId(HttpServletRequest request, String extension)
  {
    String idString = request.getParameter("id");
    Cache.info("REMOVEME setting extension=" + extension);
    Cache.info("REMOVEME idString=" + idString);
    if (idString == null)
    {
      setIdExtension(extension);
      Cache.info("REMOVEME id now " + getId());
    }
    else
    {
      setId(idString);
    }
    return getId();
  }

  protected void changeStatus(Status status)
  {
    String id = getId();
    // don't change a job's status if it has finished or died
    if (getStatus() == Status.FINISHED || getStatus() == Status.ERROR)
      return;
    api.getStatusMap().put(id, status);
  }

  protected Status getStatus()
  {
    return api.getStatusMap().get(getId());
  }

  protected void returnStatus(HttpServletResponse response)
  {
    String id = getId();
    try
    {
      PrintWriter writer = response.getWriter();
      if (id != null)
      {
        writer.write("id=" + id + "\n");
        Cache.info("REMOVEME returnStatus id=" + id);
      }
      if (api.getRequestMap().get(id) != null)
      {
        writer.write(
                "request=" + api.getRequestMap().get(id).toString() + "\n");
        Cache.info("REMOVEME returnStatus request="
                + api.getRequestMap().get(id).toString());
      }
      if (getStatus() != null)
      {
        if (getStatus() == Status.ERROR)
        {
          response.setStatus(500);
          Cache.info("REMOVEME returnStatus setting HTTP status to " + 500);
        }
        writer.write("status=" + getStatus().toString() + "\n");
        Cache.info(
                "REMOVEME returnStatus status=" + getStatus().toString());
      }
      if (getCompletableFuture() != null)
      {
        boolean isDone = getCompletableFuture().isDone();
        boolean isError = getCompletableFuture().isCompletedExceptionally();
        writer.write("cf.isDone=" + isDone + "\n");
        writer.write("cf.isError=" + isError + "\n");
        Cache.info("REMOVEME returnStatus cf.isDone=" + isDone);
        Cache.info("REMOVEME returnStatus cf.isError=" + isDone);
      }
    } catch (IOException e)
    {
      Cache.debug(e);
    }
  }

  protected boolean checkStatus(HttpServletRequest request,
          HttpServletResponse response)
  {
    return checkStatus(request, response, null);
  }

  protected boolean checkStatus(HttpServletRequest request,
          HttpServletResponse response, Status set)
  {
    String id = getId();
    Cache.info("REMOVEME checkStatus id=" + id);
    Status status = getStatus();
    if (status == null)
    {
      Cache.info("REMOVEME checkStatus status is " + null);
      if (set != null)
        changeStatus(set);
      api.getRequestMap().put(id, request.getRequestURI());
      return false;
    }
    else
    {
      Cache.info("REMOVEME checkStatus status=" + status.toString());
      return true;
    }
  }

  protected void finaliseCompletableFuture()
  {
    String id = getId();
    Cache.info("######################REMOVEME setting whenComplete on cf");
    cf.whenComplete((Void, e) -> {
      if (e != null)
      {
        Cache.error("Endpoint job " + id + " did not complete");
        Cache.debug(e);
        changeStatus(Status.ERROR);
      }
      else
      {
        Cache.info("Endpoint job " + id + " completed successfully");
        changeStatus(Status.FINISHED);
        atEnd();
      }
    });
  }
}
